import json
from unittest import TestCase

import httpx


class SecurityTest(TestCase):
    """
    Sanic Securty unit tests.
    """

    def test_authentication(self):
        """
        Registers a new account then attempts account verification and then logs into it.
        """
        client = httpx.Client()
        register_response = client.post(
            "http://127.0.0.1:8000/api/test/auth/register",
            data={
                "username": "test",
                "email": "auth@test.com",
                "password": "testtest",
            },
        )
        assert register_response.status_code == 200, register_response.text
        code = json.loads(register_response.text)["data"]
        verify_response = client.post(
            "http://127.0.0.1:8000/api/auth/verify", data={"code": code}
        )
        assert verify_response.status_code == 200, verify_response.text
        login_response = client.post(
            "http://127.0.0.1:8000/api/test/auth/login",
            data={"email": "auth@test.com", "password": "testtest"},
        )
        assert login_response.status_code == 200, login_response.text

    def test_captcha(self):
        """
        Requests a captcha and image, then attempts the captcha using the captcha solution provided in the request captcha response.
        """
        client = httpx.Client()
        account_creation_response = client.post(
            "http://127.0.0.1:8000/api/test/account/create",
            data={"email": "captcha@test.com"},
        )
        assert (
            account_creation_response.status_code == 200
        ), account_creation_response.text
        login_response = client.post(
            "http://127.0.0.1:8000/api/test/auth/login",
            data={"email": "captcha@test.com", "password": "testtest"},
        )
        assert login_response.status_code == 200, login_response.text
        captcha_request_response = client.post(
            "http://127.0.0.1:8000/api/test/capt/request"
        )
        assert (
            captcha_request_response.status_code == 200
        ), captcha_request_response.text
        captcha_img_response = client.get("http://127.0.0.1:8000/api/capt/img")
        assert captcha_img_response.status_code == 200, captcha_img_response.text
        captcha = json.loads(captcha_request_response.text)["data"]
        captcha_attempt_response = client.post(
            "http://127.0.0.1:8000/api/test/capt/attempt",
            data={"captcha": captcha},
        )
        assert (
            captcha_attempt_response.status_code == 200
        ), captcha_attempt_response.text

    def test_verification(self):
        """
        Requests two-step verification, then attempts verification using the code provided in the request verification response.
        """
        client = httpx.Client()
        account_creation_response = client.post(
            "http://127.0.0.1:8000/api/test/account/create",
            data={"email": "verification@test.com"},
        )
        assert (
            account_creation_response.status_code == 200
        ), account_creation_response.text
        verification_request_response = client.post(
            "http://127.0.0.1:8000/api/test/verif/request",
            data={"email": "verification@test.com"},
        )
        assert (
            verification_request_response.status_code == 200
        ), verification_request_response.text
        code = json.loads(verification_request_response.text)["data"]
        verification_attempt_response = client.post(
            "http://127.0.0.1:8000/api/test/verif/attempt",
            data={"code": code},
        )
        assert (
            verification_attempt_response.status_code == 200
        ), verification_attempt_response.text

    def test_role_authorization(self):
        """
        Assigns roles to the test account, then attempts to authorise this account with those roles.
        """
        client = httpx.Client()
        account_creation_response = client.post(
            "http://127.0.0.1:8000/api/test/account/create",
            data={"email": "roleauth@test.com"},
        )
        assert (
            account_creation_response.status_code == 200
        ), account_creation_response.text
        login_response = client.post(
            "http://127.0.0.1:8000/api/test/auth/login",
            data={"email": "roleauth@test.com", "password": "testtest"},
        )
        assert login_response.status_code == 200, login_response.text
        roles_assign_response = client.post(
            "http://127.0.0.1:8000/api/test/auth/roles/assign"
        )
        assert roles_assign_response.status_code == 200, roles_assign_response.text
        role_authorization_response = client.post(
            "http://127.0.0.1:8000/api/test/auth/roles/permit"
        )
        assert (
            role_authorization_response.status_code == 200
        ), role_authorization_response.text

    def test_permission_authorization(self):
        """
        Assigns permissions to the test account, then attempts to authorise this account with those permissions.
        """
        client = httpx.Client()
        account_creation_response = client.post(
            "http://127.0.0.1:8000/api/test/account/create",
            data={"email": "permauth@test.com"},
        )
        assert (
            account_creation_response.status_code == 200
        ), account_creation_response.text
        login_response = client.post(
            "http://127.0.0.1:8000/api/test/auth/login",
            data={"email": "permauth@test.com", "password": "testtest"},
        )
        assert login_response.status_code == 200, login_response.text
        permissions_assign_response = client.post(
            "http://127.0.0.1:8000/api/test/auth/perms/assign"
        )
        assert (
            permissions_assign_response.status_code == 200
        ), permissions_assign_response.text
        permission_authorization_response = client.post(
            "http://127.0.0.1:8000/api/test/auth/perms/permit"
        )
        assert (
            permission_authorization_response.status_code == 200
        ), permission_authorization_response.text

    def test_recovery(self):
        """
        Requests password recovery for a newly created test account, then attempts to recover password using the code found in the recovery request response.
        Once the password is changed, a login attempt is made to the test account with the new password.
        """
        client = httpx.Client()
        account_creation_response = client.post(
            "http://127.0.0.1:8000/api/test/account/create",
            data={"email": "recovery@test.com"},
        )
        assert (
            account_creation_response.status_code == 200
        ), account_creation_response.text
        recovery_request_response = client.post(
            "http://127.0.0.1:8000/api/test/recov/request",
            data={"email": "recovery@test.com"},
        )
        assert (
            recovery_request_response.status_code == 200
        ), recovery_request_response.text
        code = json.loads(recovery_request_response.text)["data"]
        recovery_recover_response = client.post(
            "http://127.0.0.1:8000/api/recov/recover",
            data={"password": "recovered", "code": code},
        )
        assert (
            recovery_recover_response.status_code == 200
        ), recovery_recover_response.text
        login_response = client.post(
            "http://127.0.0.1:8000/api/test/auth/login",
            data={"email": "recovery@test.com", "password": "recovered"},
        )
        assert login_response.status_code == 200, login_response.text



from sanic.exceptions import SanicException

from sanic_security.utils import json


class SecurityError(SanicException):
    """
    Sanic Security related error.

    Attributes:
        response (HTTPResponse): Security Error json response.

    Args:
        message (str): Human readable error message.
        code (int): HTTP error code.
    """

    def __init__(self, message: str, code: int):
        self.response = json(message, self.__class__.__name__, code)
        super().__init__(message, code)


class NotFoundError(SecurityError):
    def __init__(self, message):
        super().__init__(message, 404)


class DeletedError(SecurityError):
    def __init__(self, message):
        super().__init__(message, 404)


class AccountError(SecurityError):
    def __init__(self, message, code):
        super().__init__(message, code)


class ExistsError(AccountError):
    def __init__(self):
        super().__init__("Account with this email or phone number already exists.", 409)


class InvalidIdentifierError(AccountError):
    def __init__(self, message):
        super().__init__(message, 400)


class DisabledError(AccountError):
    def __init__(self):
        super().__init__("This account has been disabled.", 401)


class PasswordIncorrectError(AccountError):
    def __init__(self):
        super().__init__("The password provided is incorrect.", 401)


class UnverifiedError(AccountError):
    def __init__(self):
        super().__init__("Account requires verification.", 401)


class SessionError(SecurityError):
    """
    A session related error.

    Args:
        message (str): Human readable error message.
        code (int): HTTP Error code.
    """

    def __init__(self, message, code):
        super().__init__(message, code)


class DecodingError(SessionError):
    def __init__(self, exception):
        super().__init__(f"Session could not be decoded. {exception}", 400)


class InvalidError(SessionError):
    def __init__(self):
        super().__init__("Session is invalid.", 401)


class ExpiredError(SessionError):
    def __init__(self):
        super().__init__("Session has expired", 401)


class CrosscheckError(SessionError):
    def __init__(self, message="The code provided is incorrect."):
        super().__init__(message, 401)


class MaximumAttemptsError(SessionError):
    def __init__(self):
        super().__init__(
            "You've reached the maximum amount of attempts for this session.", 401
        )


class InsufficientRoleError(SecurityError):
    def __init__(self):
        super().__init__("Insufficient roles required for this action.", 403)


class InsufficientPermissionError(SecurityError):
    def __init__(self):
        super().__init__("Insufficient permissions required for this action.", 403)



from sanic.request import Request

from sanic_security.models import (
    AuthenticationSession,
    Account,
    TwoStepSession,
    AccountErrorFactory,
    SessionFactory,
)
from sanic_security.utils import hash_password

account_error_factory = AccountErrorFactory()
session_factory = SessionFactory()


async def recover_password(request: Request, two_step_session: TwoStepSession):
    """
    Recovers an account by changing the password once recovery request was determined to be made by the account owner
    via requiring two-step verification.

    Args:
        request (Request): Sanic request parameter. All request bodies are sent as form-data with the following arguments: password.
        two_step_session (TwoStepSession): Two-step session containing account being recovered.

    """
    two_step_session.account.password = hash_password(request.form.get("password"))
    await AuthenticationSession.filter(
        account=two_step_session.account, valid=True, deleted=False
    ).update(valid=False)
    await two_step_session.account.save(update_fields=["password"])


async def request_password_recovery(request: Request):
    """
    Requests a two-step session to ensure that recovery request was made by the account owner.

    Args:
        request (Request): Sanic request parameter. All request bodies are sent as form-data with the following arguments: email.

    Returns:
        two_step_session
    """

    account = await Account.get_via_email(request.form.get("email"))
    account_error_factory.throw(account)
    return await session_factory.get("twostep", request, account=account)



import asyncio
import datetime
import os
import random
import string
import uuid
import aiofiles
import jwt
from captcha.image import ImageCaptcha
from jwt import DecodeError
from sanic import Sanic
from sanic.request import Request
from sanic.response import HTTPResponse, file
from tortoise import fields, Model

from sanic_security.exceptions import *
from sanic_security.utils import get_ip, security_cache_path, dir_exists, config
from sanic_security.lib.smtp import send_email
from sanic_security.lib.twilio import send_sms


class BaseErrorFactory:
    """
    Easily raise or retrieve errors based off of model variable values.
    """

    def get(self, model):
        """
        Retrieves an error based off of defined variable conditions.

        Args:
            model: Model being used to check for validation.

        Raises:
            SecurityError
        """
        raise NotImplementedError()

    def throw(self, model):
        """
        Raises an error that was retrieved in the get method.

        Args:
            model (BaseModel): Model being used to check for validation.

        Raises:
            SecurityError
        """
        error = self.get(model)
        if error:
            raise error


class BaseModel(Model):
    """
    Base Sanic Security model that all other models derive from.

    Attributes:
        id (int): Primary key of model.
        uid (bytes): Recommended identifier to be used for filtering or retrieval.
        account (Account): Parent account associated with this model.
        date_created (datetime): Time this model was created in the database.
        date_updated (datetime): Time this model was updated in the database.
        deleted (bool): This attribute allows you to mark a model as deleted and filter it from queries without removing it from the database.
    """

    id = fields.IntField(pk=True)
    uid = fields.UUIDField(unique=True, default=uuid.uuid1, max_length=36)
    account = fields.ForeignKeyField("models.Account", null=True)
    date_created = fields.DatetimeField(auto_now_add=True)
    date_updated = fields.DatetimeField(auto_now=True)
    deleted = fields.BooleanField(default=False)

    def json(self):
        """
        Retrieve a JSON serializable dict to be used in a HTTP request or response.

        Example:
            Below is an example of this method returning a dict to be used for JSON serialization.

                def json(self):
                    return {
                        'uid': str(self.uid),
                        'date_created': str(self.date_created),
                        'date_updated': str(self.date_updated),
                        'email': self.email,
                        'username': self.username,
                        'disabled': self.disabled,
                        'verified': self.verified
                    }

        """
        raise NotImplementedError()

    class Meta:
        abstract = True


class Account(BaseModel):
    """
    Contains all identifiable user information.

    Attributes:
        username (str): Public identifier.
        email (str): Private identifier and can be used for verification.
        phone (str): Mobile phone number with country code included and can be used for verification.
        password (bytes): Must be created using the hash_password method found in the utils module.
        disabled (bool): Renders an account unusable but available for moderators to investigate for infractions.
        verified (bool): Determines if an account has been through the two-step verification process before being allowed use.
    """

    username = fields.CharField(max_length=45)
    email = fields.CharField(unique=True, max_length=45)
    phone = fields.CharField(unique=True, max_length=14, null=True)
    password = fields.BinaryField()
    disabled = fields.BooleanField(default=False)
    verified = fields.BooleanField(default=False)

    def json(self):
        return {
            "uid": str(self.uid),
            "date_created": str(self.date_created),
            "date_updated": str(self.date_updated),
            "email": self.email,
            "username": self.username,
            "disabled": self.disabled,
            "verified": self.verified,
        }

    @staticmethod
    async def get_via_email(email: str):
        """
        Retrieve an account with an email.

        Args:
            email (str): Email associated to account being retrieved.

        Returns:
            account
        """
        account = await Account.filter(email=email).first()
        return account


class AccountErrorFactory(BaseErrorFactory):
    def get(self, model):
        error = None
        if not model:
            error = NotFoundError("This account does not exist.")
        elif model.deleted:
            error = DeletedError("This account has been permanently deleted.")
        elif model.disabled:
            error = DisabledError()
        elif not model.verified:
            error = UnverifiedError()
        return error


class Session(BaseModel):
    """
    Used for client identification and validation. Base session model that all session models derive from.

    Attributes:
        expiration_date (datetime): Time the session expires and can no longer be used.
        valid (bool): Used to determine if a session can be utilized or not.
        ip (str): IP address of client creating session.
        cache_path (str): Session cache path.
    """

    expiration_date = fields.DatetimeField(null=True)
    valid = fields.BooleanField(default=True)
    ip = fields.CharField(max_length=16)
    cache_path = "./resources/security-cache/session/"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.cookie = f"{config['SECURITY']['name']}_{self.__class__.__name__}"

    def json(self):
        return {
            "uid": str(self.uid),
            "date_created": str(self.date_created),
            "date_updated": str(self.date_updated),
            "expiration_date": str(self.expiration_date),
            "account": self.account.email
            if isinstance(self.account, Account)
            else None,
            "valid": self.valid,
        }

    def encode(self, response: HTTPResponse, secure=True, same_site: str = "lax"):
        """
        Transforms session into jwt and then is stored in a cookie.

        Args:
            secure (bool): If true, connections must be SSL encrypted (aka https).
            response (HTTPResponse): Sanic response object used to store JWT into a cookie on the client.
            same_site (bool): : Allows you to declare if your cookie should be restricted to a first-party or same-site context.
        """
        payload = {
            "date_created": str(self.date_created),
            "expiration_date": str(self.expiration_date),
            "uid": str(self.uid),
            "ip": self.ip,
        }
        encoded = jwt.encode(payload, config["SECURITY"]["secret"], "HS256")
        response.cookies[self.cookie] = encoded
        response.cookies[self.cookie]["secure"] = secure
        response.cookies[self.cookie]["samesite"] = same_site

    def decode_raw(self, request: Request):
        """
        Decodes JWT token from client cookie into a python dict.

        Args:
            request (Request): Sanic request parameter.

        Returns:
            session_dict
        """
        cookie = request.cookies.get(self.cookie)
        try:
            if not cookie:
                raise DecodingError(f"No session provided by client.")
            else:
                return jwt.decode(cookie, config["SECURITY"]["secret"], "HS256")
        except DecodeError as e:
            raise DecodingError(e)

    async def decode(self, request: Request):
        """
        Decodes JWT token from client cookie to a Sanic Security session.

        Args:
            request (Request): Sanic request parameter.

        Returns:
            session

        Raises:
            DecodeError
        """
        decoded = self.decode_raw(request)
        return (
            await self.filter(uid=decoded.get("uid"))
            .prefetch_related("account")
            .first()
        )

    async def crosscheck_location(self, request):
        """
        Checks if client using session is in a known location (ip address). Prevents cookie jacking.

        Raises:
            UnknownLocationError:
        """

        if not await self.filter(ip=get_ip(request), account=self.account).exists():
            raise CrosscheckError(
                "Client location does not match any existing session location."
            )

    class Meta:
        abstract = True


class SessionErrorFactory(BaseErrorFactory):
    def get(self, model):
        error = None
        if model is None:
            error = NotFoundError("Session could not be found.")
        elif not model.valid:
            error = InvalidError()
        elif model.deleted:
            error = DeletedError("Session has been deleted.")
        elif datetime.datetime.now(datetime.timezone.utc) >= model.expiration_date:
            error = ExpiredError()
        return error


class VerificationSession(Session):
    """
    Used for a client verification method that requires some form of code, challenge, or key.

    Attributes:
        attempts (int): The amount of incorrect times a user entered a code not equal to this verification sessions code.
        code (str): Used as a secret key that would be sent or provided in a way that makes it difficult for malicious actors to obtain.
    """

    attempts = fields.IntField(default=0)
    code = fields.CharField(max_length=10, null=True)

    @staticmethod
    async def initialize_cache():
        """
        Creates verification session cache and generates required files.

        Args:
            app (Sanic): Sanic Framework app.
        """
        raise NotImplementedError()

    @classmethod
    async def get_random_code(cls):
        """
        Retrieves a random cached verification session code.
        """
        raise NotImplementedError()

    async def crosscheck_code(self, code: str):
        """
        Used to check if code passed is equivalent to the verification session code.

        Args:
            code (str): Code being cross-checked with verification session code.

        Raises:
            CrossCheckError
            MaximumAttemptsError
        """
        if self.attempts >= 5:
            raise MaximumAttemptsError()
        elif self.code != code:
            self.attempts += 1
            await self.save(update_fields=["attempts"])
            raise CrosscheckError("The value provided does not match.")
        else:
            self.valid = False
            await self.save(update_fields=["valid"])

    class Meta:
        abstract = True


class TwoStepSession(VerificationSession):
    """
    Validates a client using a code sent via email or text.
    """

    @staticmethod
    async def initialize_cache():
        if not dir_exists(f"{security_cache_path}/verification"):
            async with aiofiles.open(
                f"{security_cache_path}/verification/codes.txt", mode="w"
            ) as f:
                for i in range(100):
                    code = "".join(
                        random.choices(string.ascii_letters + string.digits, k=10)
                    )
                    await f.write(code + " ")

    @classmethod
    async def get_random_code(cls):
        await cls.initialize_cache()
        async with aiofiles.open(
            f"{security_cache_path}/verification/codes.txt", mode="r"
        ) as f:
            codes = await f.read()
            return random.choice(codes.split())

    async def text_code(self, code_prefix="Your code is: "):
        """
        Sends account associated with this session the code via text.

        Args:
            code_prefix (str): Message being sent with code, for example "Your code is: ".
        """
        await send_sms(self.account.phone, code_prefix + self.code)

    async def email_code(
        self, subject="Verification", code_prefix="Your code is:\n\n "
    ):
        """
        Sends account associated with this session the code via email.

        Args:
            code_prefix (str): Message being sent with code, for example "Your code is: ".
            subject (str): Subject of email being sent with code.
        """
        await send_email(self.account.email, subject, code_prefix + self.code)


class CaptchaSession(VerificationSession):
    """
    Validates a client as human with a captcha challenge.
    """

    @staticmethod
    async def initialize_cache():
        if not dir_exists(f"{security_cache_path}/captcha"):
            loop = asyncio.get_running_loop()
            image = ImageCaptcha(190, 90, fonts=[config["SECURITY"]["captcha_font"]])
            for i in range(100):
                code = "".join(
                    random.choices("123456789qQeErRtTyYiIaAdDfFgGhHkKlLbBnN", k=6)
                )
                await loop.run_in_executor(
                    None,
                    image.write,
                    code,
                    f"{security_cache_path}/captcha/{code}.png",
                )

    @classmethod
    async def get_random_code(cls):
        await cls.initialize_cache()
        return random.choice(os.listdir(f"{security_cache_path}/captcha")).split(".")[0]

    async def get_image(self):
        """
        Retrieves captcha image file.

        Returns:
            captcha_image
        """
        try:
            captcha_image_response = await file(
                f"{security_cache_path}/captcha/{self.code}.png"
            )
        except Exception as e:
            captcha_image_response = json(
                "Could not retrieve captcha image.", str(e), e.__name__
            )
        return captcha_image_response


class AuthenticationSession(Session):
    """
    Used to authenticate a client and provide access to a user's account.
    """

    pass


class SessionFactory:
    """
    Prevents human error when creating sessions.
    """

    def generate_expiration_date(self, days: int = 0, minutes: int = 0):
        """
        Creates an expiration date. Adds days to current datetime.

        Args:
            days (int):  Days to be added to current time.
            minutes (int): Minutes to be added to the current time.

        Returns:
            expiration_date
        """
        return datetime.datetime.utcnow() + datetime.timedelta(
            days=days, minutes=minutes
        )

    async def get(self, session_type: str, request: Request, account: Account = None):
        """
         Creates and returns a session with all of the fulfilled requirements.

        Args:
            session_type (str): The type of session being retrieved. Available types are: captcha, twostep, and authentication.
            request (Request):  Sanic request paramater.
            account (Account): Account being associated to a session.

        Returns:
            session

        Raises:
            ValueError
        """
        if session_type == "captcha":
            return await CaptchaSession.create(
                ip=get_ip(request),
                code=await CaptchaSession.get_random_code(),
                expiration_date=self.generate_expiration_date(minutes=1),
            )
        elif session_type == "twostep":
            return await TwoStepSession.create(
                code=await TwoStepSession.get_random_code(),
                ip=get_ip(request),
                account=account,
                expiration_date=self.generate_expiration_date(minutes=5),
            )
        elif session_type == "authentication":
            return await AuthenticationSession.create(
                account=account,
                ip=get_ip(request),
                expiration_date=self.generate_expiration_date(days=30),
            )
        else:
            raise ValueError("Invalid session type.")


class Role(BaseModel):
    """
    Assigned to an account to authorize an action. Used for role based authorization.

    Attributes:
        name (str): Name of the role.
    """

    name = fields.CharField(max_length=45)

    def json(self):
        return {
            "uid": str(self.uid),
            "date_created": str(self.date_created),
            "date_updated": str(self.date_updated),
            "name": self.name,
        }


class Permission(BaseModel):
    """
    Assigned to an account to authorize an action. Used for wildcard based authorization.

    Attributes:
        wildcard (str): The wildcard for this permission.
    """

    wildcard = fields.CharField(max_length=45)

    def json(self):
        return {
            "uid": str(self.uid),
            "date_created": str(self.date_created),
            "date_updated": str(self.date_updated),
            "wildcard": self.wildcard,
        }



import functools

from sanic import Request

from sanic_security.models import CaptchaSession, SessionFactory, SessionErrorFactory

session_factory = SessionFactory()
session_error_factory = SessionErrorFactory()


async def request_captcha(request: Request):
    """
    Creates a captcha session associated with an account.

    Args:
        request (Request): Sanic request parameter.

    Returns:
        captcha_session
    """
    return await session_factory.get("captcha", request)


async def captcha(request: Request):
    """
    Verifies a captcha challenge attempt.

    Args:
        request (Request): Sanic request parameter. All request bodies are sent as form-data with the following arguments: captcha.

    Raises:
        SessionError

    Returns:
        captcha_session
    """
    captcha_session = await CaptchaSession().decode(request)
    session_error_factory.throw(captcha_session)
    await captcha_session.crosscheck_code(request.form.get("captcha"))
    return captcha_session


def requires_captcha():
    """
    Verifies a captcha attempt.

    Example:
        This method is not called directly and instead used as a decorator:

            @app.post("api/captcha/attempt")
            @requires_captcha()
            async def on_captcha_attempt(request, captcha_session):
                return json("Captcha attempt successful!", captcha_session.json())
    Raises:
        SessionError
    """

    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(request, *args, **kwargs):
            captcha_session = await captcha(request)
            return await func(request, captcha_session, *args, **kwargs)

        return wrapped

    return wrapper



from sanic import Blueprint

from sanic_security.authentication import (
    login,
    logout,
    register,
    requires_authentication,
)
from sanic_security.captcha import requires_captcha, request_captcha
from sanic_security.models import CaptchaSession, TwoStepSession
from sanic_security.recovery import (
    request_password_recovery,
    recover_password,
)
from sanic_security.utils import json
from sanic_security.verification import (
    request_two_step_verification,
    requires_two_step_verification,
    verify_account,
)

authentication = Blueprint("authentication_blueprint")
verification = Blueprint("verification_blueprint")
recovery = Blueprint("recovery_blueprint")
captcha = Blueprint("captcha_blueprint")
security = Blueprint.group(authentication, verification, recovery, captcha)


@authentication.post("api/auth/register")
@requires_captcha()
async def on_register(request, captcha_session):
    """
    Register an account with an email, username, and password. Once the account is created successfully, a two-step session is requested and the code is emailed.
    """
    two_step_session = await register(request)
    await two_step_session.email_code()
    response = json("Registration successful!", two_step_session.account.json())
    two_step_session.encode(response)
    return response


@authentication.post("api/auth/login")
async def on_login(request):
    """
    Login with an email and password.
    """
    authentication_session = await login(request)
    response = json("Login successful!", authentication_session.account.json())
    authentication_session.encode(response)
    return response


@authentication.post("api/auth/verify")
@requires_two_step_verification()
async def on_verify(request, two_step_session):
    """
    Verify account with a two-step session code found in email.
    """
    await verify_account(two_step_session)
    return json("Account verification successful!", two_step_session.account.json())


@authentication.post("api/auth/logout")
@requires_authentication()
async def on_logout(request, authentication_session):
    """
    Logout of logged in account.
    """
    await logout(authentication_session)
    response = json("Logout successful!", authentication_session.account.json())
    return response


@verification.post("api/verif/resend")
async def on_resend_verification(request):
    """
    Resend existing two-step session code if lost.
    """
    two_step_session = await TwoStepSession().decode(request)
    await two_step_session.email_code()
    response = json("Verification resend successful!", two_step_session.account.json())
    return response


@verification.post("api/verif/request")
@requires_captcha()
async def on_request_verification(request, captcha_session):
    """
    Request new two-step session and send email with code.
    """
    two_step_session = await request_two_step_verification(request)
    await two_step_session.email_code()
    response = json("Verification request successful!", two_step_session.json())
    two_step_session.encode(response)
    return response


@recovery.post("api/recov/request")
@requires_captcha()
async def on_recovery_request(request, captcha_session):
    """
    Requests new two-step session to ensure current recovery attempt is being made by account owner.
    """
    two_step_session = await request_password_recovery(request)
    await two_step_session.email_code()
    response = json("Recovery request successful!", two_step_session.account.json())
    two_step_session.encode(response)
    return response


@recovery.post("api/recov/recover")
@requires_two_step_verification()
async def on_recover(request, two_step_session):
    """
    Changes an account's password once recovery attempt was determined to have been made by account owner with two-step code found in email.
    """
    await recover_password(request, two_step_session)
    return json("Account recovered successfully", two_step_session.account.json())


@captcha.post("api/capt/request")
async def on_request_captcha(request):
    """
    Requests new captcha session.
    """
    captcha_session = await request_captcha(request)
    response = json("Captcha request successful!", captcha_session.json())
    captcha_session.encode(response)
    return response


@captcha.get("api/capt/img")
async def on_captcha_img_request(request):
    """
    Retrieves captcha image from existing captcha session.
    """
    captcha_session = await CaptchaSession().decode(request)
    return await captcha_session.get_image()



from sanic import Sanic
from tortoise import Tortoise

from sanic_security.utils import config


def initialize_security_orm(app: Sanic):
    """
    Initializes tortoise-orm.

    Args:
        app (Sanic): Sanic Framework app.
    """

    @app.listener("before_server_start")
    async def init_orm(app, loop):
        username = config["TORTOISE"]["username"]
        password = config["TORTOISE"]["password"]
        endpoint = config["TORTOISE"]["endpoint"]
        schema = config["TORTOISE"]["schema"]
        engine = config["TORTOISE"]["engine"]
        models = config["TORTOISE"]["models"].replace(" ", "").split(",")
        url = f"{engine}://{username}:{password}@{endpoint}/{schema}"
        await Tortoise.init(db_url=url, modules={"models": models})
        if config["TORTOISE"]["generate"] == "true":
            await Tortoise.generate_schemas()

    @app.listener("after_server_stop")
    async def close_orm(app, loop):
        await Tortoise.close_connections()
